#!/usr/bin/env python
# -*- coding: utf-8 -*-

# @Date: 2022/8/12
# @Author: pcl
from .modelinfo.modelinfo import ModelInfo
from .infer.infer import Infer

__all__ = [
    "ModelInfo",
    "Infer",
]
